import pyseafuel._constants as _constants
import numpy as np

# ──────────────────────────────────────────────────────────────────────────

def ideal(volume_seawater_in, S, T, water_ratio, efficiency=1, brine_concentration=False):
    """
    Thermodynamic limit for desalinating an inflow of saline water to produce pure water[1]_.

    Parameters
    ----------
    volume_seawater_in : float
        Seawater inflow; liters per second.
    S : float
        Seawater inflow salinity; Practical Salinity Units.
    T : float
        Temperature of the seawater inflow; Kelvin.
    water_ratio : float
        Ratio of the inflow separated into pure water; between [0, 1].
    P : float (optional)
        Operating pressure of the inflow; decibars; default = 10.1325 db.
    efficiency : float (optional)
        Efficiency of the desalination process; between [0, 1]; default = 1.
    brine_concentration : boolean (optional)
        Returns brine molar concentration if set to True; default = False.

    Returns
    -------
    volume_h2o_out : float
        Flow of pure water out, in liters per second.
    volume_brine_out : float
        Flow of brine out, in liters per second.
    power_consumed : float
        Power consumed by the desalination process, in Watts.
    c_b : float (optional)
        Brine molar concentration, in moles per liter.

    References
    ----------
    .. [1] `L. Wang, C. Violet, R. DuChanois, and M. Elimelech, "Derivation of the Theoretical Minimum Energy of Separation of Desalination Process," Journal of Chemical Education, vol. 97, pp. 4361-4369, 2020.<https://doi.org/10.1021/acs.jchemed.0c01194>`_
    """

    # molar concentration of salt in inflow
    ppt = S  # PSU approxiamtely equal to parts per thousand (PPT); which is essentially in g/L;
    c_f = ppt / _constants.mol_weight_nacl / 1000  # mol/L

    pi_f = 2 * _constants.R * T * c_f   # J/L; osmotic pressure of sea water

    # calculating power consumed
    sec = -pi_f / water_ratio * np.log(1 - water_ratio)  # J/L; sec is the specific energy consumption to produce the pure water given the saline inflow
    power_consumed = volume_seawater_in * sec * efficiency  # W; power required to input feed 

    # output flows
    volume_h2o_out = water_ratio * volume_seawater_in  # L/s; output of pure water
    volume_brine_out = (1 - water_ratio) * volume_seawater_in  # L/s; output of brine

    c_b = c_f / (1 - water_ratio)  # brine concentration; moles per liter

    if brine_concentration:
        return volume_h2o_out, volume_brine_out, power_consumed, c_b

    return volume_h2o_out, volume_brine_out, power_consumed


def reverse_osmosis(volume_seawater_in, S, T, water_ratio, stages, brine_concentration=False):
    """
    Multistage reverse osmosis desalinator for desalinating an inflow of saline water to produce pure water[1]_.

    Parameters
    ----------
    volume_seawater_in : float
        Seawater inflow; liters per second.
    S : float
        Seawater inflow salinity; Practical Salinity Units.
    T : float
        Temperature of the seawater inflow; Kelvin.
    water_ratio : float
        Ratio of the inflow separated into pure water; between [0, 1].
    stages : int
        Number of stages.
    brine_concentration : boolean (optional)
        Returns brine molar concentration if set to True; default = False.

    Returns
    -------
    volume_h2o_out : float
        Flow of pure water out; liters per second.
    volume_brine_out : float
        Flow of brine out; liters per second.
    power_consumed : float
        Power consumed by the desalination process; Watts.
    c_b : float (optional)
        Brine molar concentration; moles per liter.

    References
    ----------
    .. [1] `L. Wang, C. Violet, R. DuChanois, and M. Elimelech, "Derivation of the Theoretical Minimum Energy of Separation of Desalination Process," Journal of Chemical Education, vol. 97, pp. 4361-4369, 2020.<https://doi.org/10.1021/acs.jchemed.0c01194>`_
    """

    def _water_ratio_i(water_ratio, stages, i):
        """Water ratio per stage.
        """

        water_ratio_i = water_ratio * i / ((1 - water_ratio) * stages + water_ratio * i)

        return water_ratio_i

    # molar concentration of salt in the inflow
    ppt = S  # PSU approxiamtely equal to parts per thousand (PPT); which is essentially in g/L;
    c_f = ppt / _constants.mol_weight_nacl / 1000  # mol/L

    pi_f = 2 * _constants.R * T * c_f  # J/l; osmotic pressure of sea water

    # calculating specific energy consumption
    sec = 0

    for i in np.arange(1, stages+1):
        sec += (_water_ratio_i(water_ratio, stages, i) - _water_ratio_i(water_ratio, stages, i-1)) / (1 - _water_ratio_i(water_ratio, stages, i))

    sec *= pi_f / water_ratio  # J/L; sec is the specific energy consumption to produce the pure water given the saline inflow

    # power consumed
    power_consumed = volume_seawater_in * sec  # W; power required to input feed 

    # volume outflows
    volume_h2o_out = water_ratio * volume_seawater_in  # L/s; output of pure water
    volume_brine_out = (1 - water_ratio) * volume_seawater_in  # L/s; output of brine

    c_b = c_f / (1 - water_ratio)  # mol/L; brince concentration

    if brine_concentration:
        return volume_h2o_out, volume_brine_out, power_consumed, c_b

    return volume_h2o_out, volume_brine_out, power_consumed


def electrodialysis(volume_seawater_in, S, T, water_ratio, salt_removal, stages, brine_concentration=False):
    """
    Multistage electrodialysis desalinator for desalinating an inflow of saline water to produce pure water[1]_.

    Parameters
    ----------
    volume_seawater_in : float
        Seawater inflow; liters per second.
    S : float
        Seawater inflow salinity; Practical Salinity Units.
    T : float
        Seawater inflow temperature; Kelvin.
    water_ratio : float
        Ratio of the inflow separated into pure water; between [0, 1].
    salt_removal : float
        Percentage of salt removal in terms of concentration for the output flow; between (0, 1).
    stages : int
        Number of stages.
    brine_concentration : boolean (optional)
        Returns brine molar concentration if set to True; default = False.

    Returns
    -------
    volume_h2o_out : float
        Flow of pure water out; liters per second.
    volume_brine_out : float
        Flow of brine out; liters per second.
    power_consumed : float
        Power consumed by the desalination process; Watts.
    c_b : float (optional)
        Brine molar concentration; moles per liter.

    References
    ----------
    .. [1] `L. Wang, C. Violet, R. DuChanois, and M. Elimelech, "Derivation of the Theoretical Minimum Energy of Separation of Desalination Process," Journal of Chemical Education, vol. 97, pp. 4361-4369, 2020.<https://doi.org/10.1021/acs.jchemed.0c01194>`_
    """
    
    # per stage values
    salt_removal_i = 1 - (1 - salt_removal)**(1/stages)  # salt removal percentage per stage in terms of concentration
    water_ratio_i = water_ratio**(1/stages)  # water ratio per stage

    # Calculating inflow salt molar concentration
    ppt = S  # PSU approxiamtely equal to parts per thousand (PPT); which is essentially in g/L;
    c_f = ppt / _constants.mol_weight_nacl / 1000  # mol/L

    # calculating specific energy consumption
    sec = 0
    
    for i in np.arange(1, stages+1):
        sec += (1 - salt_removal_i)**(i-1)

    sec *= 2 * _constants.R * T * salt_removal_i * c_f * np.log(1 / ( (1 - salt_removal_i) * (1 - water_ratio_i)) - water_ratio_i / (1 - water_ratio_i))

    # calculating power consumption
    power_consumed = volume_seawater_in * sec

    # volume outflows
    volume_h2o_out = water_ratio * volume_seawater_in  # L/s; output of pure water
    volume_brine_out = (1 - water_ratio) * volume_seawater_in  # L/s; output of brine

    # brine concentration
    c_b = c_f * (1 - (1 - salt_removal) * water_ratio) / (1 - water_ratio)  # mol/L; brine molar concentration

    if brine_concentration:
        return volume_h2o_out, volume_brine_out, power_consumed, c_b

    return volume_h2o_out, volume_brine_out, power_consumed



"""
This is the documentation for the python package [**pyseafuel**](https://pypi.org/project/pyseafuel/).
"""

# ──────────────────────────────────────────────────────────────────────────

__version__ = '0.1.0'

# ──────────────────────────────────────────────────────────────────────────

from pyseafuel import degasser
from pyseafuel import desalinator
from pyseafuel import dic
from pyseafuel import electrolyzer
from pyseafuel import reactor
from pyseafuel import system



import pyseafuel._constants as _constants
import numpy as np

# ──────────────────────────────────────────────────────────────────────────

def Mcphy_Piel(volume_h2o_in):
    """
    Estimating power consumption and outflows for electrolyzing pure water with an McPhy Piel series electrolyzer.

    Parameters
    ----------
    volume_h2o_in : float
        Pure water inflow; liters per second.

    Returns
    -------
    mass_h2_out : float
        Mass flow rate out of hydrogen gas; kilograms per second.
    mass_o2_out : float
        Mass flow rate out of oxygen gas; kilograms per second.
    power_consumed : float
        Power consumed during electrolysis; Watts.

    Notes
    -----
    The values to calculate the process of electrolysis come from the datasheet for the `McPhy Piel series H model electrolyzer. <https://mcphy.com/en/equipment-services/electrolyzers/small/>`_ It consumes 18 kW at an outflow of 3 m$^3$/h of hydrogen and 60 kW at 10 m$^3$/h.

    This function is only accurate if the water inflow **volume_h2o_in** remains between 6.69 x 10$^{-4}$ L/s and 2.23 x 10$^{-3}$ L/s.
    """

    volume_h2o_in /= 1000  # m^3/s; converting from liters to cubic meters

    ratio_h2_o2 = 2*_constants.rho_h2/_constants.rho_o2  # ratio of hydrogen over oxygen mass outflow rates

    # output volumes
    volume_h2_out = volume_h2o_in * _constants.rho_h2o / (1 + 1/ratio_h2_o2) / _constants.rho_h2  # m^3/s
    volume_o2_out = volume_h2_out * .5  # m^3/s

    # output masses
    mass_h2_out = volume_h2_out * _constants.rho_h2  # kg/s
    mass_o2_out = volume_o2_out * _constants.rho_o2  # kg/s

    linear_fit = np.polyfit([3/3600, 10/3600], [18, 60], 1)  # linear interpolation/extrapolation to estimate power consumed
    power_consumed = np.polyval(linear_fit, volume_h2_out) * 1000  # W

    return mass_h2_out, mass_o2_out, power_consumed


def Shen(volume_h2o_in, E0, K, R, A, output_voltage=False):
    """
    Calculates power consumption and outflows for electrolyzing pure water using the concise model from Shen et al. 2011 (see References section).

    Parameters
    ----------
    volume_h2o_in : float
        Volumetric inflow of water; liters per second.
    E0 : float
        Reversible cell potential; depends on the cell's materials; Volts.
    K : float
        Power conversion coefficient; unity per ohm squared centimeter.
    R : float
        Cell (internal) resistance; ohm squared centimeter.
    A : float
        Membrane stack area; squared centimeter.
    output_voltage : boolean
        Flag to return voltage, **V**, with other return variables.

    Returns
    -------
    mass_h2_out : float
        Mass outflow of hydrogen; kilograms per second.
    mass_o2_out : float
        Mass outflow of oxygen; kilograms per second.
    power_consumed : float
        Power consumed during electrolysis.
    V : float (optional)
        Operating voltage, $V$; see **output_voltage**.

    Notes
    -----
    Power calculation comes from the model in Shen et al. 2011[1]_. Essentially, power is assumed to be proportional to the square of the cell voltage, $V$, internal resistance, $R$, and cell dissociation potential, $E_0$:

    $$
    P = K ( V - IR - E_0)^2
    $$

    $K$ and $R$ are given in units with cm$^2$ because electrolysis processes are typically measured with current density rather than current.
    
    References
    ----------
    .. [1] `N. Shen, N. Bennett, Y. Ding, and K. Scott, "A concise model for evaluating water electrolysis," International Journal of Hydrogen Energy, vol. 36, pp. 14335-14341, 2011.<http://dx.doi.org/10.1016/j.ijhydene.2010.12.029>`_
    """

    volume_h2o_in /= 1000  # m^3/s; converting from liters to cubic meters
    mass_h2o_in = volume_h2o_in * _constants.rho_h2o  # kg/s

    # molecular flow rates
    mol_h2o = mass_h2o_in / _constants.mol_weight_h2o
    mol_h2 = mol_h2o
    mol_o2 = mol_h2o/2

    # mass outflow rates
    mass_h2_out = mol_h2 * _constants.mol_weight_h2
    mass_o2_out = mol_o2 * _constants.mol_weight_o2

    # power calculation
    I = _constants.F * 2 * mol_h2o / A  # A/cm^2; current density

    a = K
    b = K * (-2*I*R - 2*E0) - I
    c = K * (2*E0*I*R + I**2*R**2 + E0**2) + I*R**2

    V = (-b + (b**2 - 4*a*c)**(1/2)) / (2 * a)  # V; model from reference

    power_consumed = V * I * A  # W

    if output_voltage:
        return mass_h2_out, mass_o2_out, power_consumed, V

    else:
        return mass_h2_out, mass_o2_out, power_consumed



import pyseafuel
import pyseafuel._constants as _constants

# ──────────────────────────────────────────────────────────────────────────

def seafuel(degasser, desalinator, electrolyzer, reactor):
    """

    Parameters
    ----------
    degasser : dict
        Degasser information; must contain the 'type' key. 'type' can equal on of the following: 'bipolar_membrane'.
        'bipolar_membrane' requires 'seawater_in' (seawater inflow) in L/s.
    desalintor : dict
        Desalinator information; must contain the 'type' key. 'type' can equal on of the following: 'electrodialysis'.
        'electrodialysis' requires 'seawater_in' (seawater inflow) in L/s, 'seawater_S' (seawater salinity) in PSU, 'seawater_T' (seawater temperature) in Kelvin, 'water_ratio' (volume ratio of fresh water out over seawater in) from (0, 1), 'salt_removal' (fraction of salt removed from the fresh water outflow) from (0, 1), and 'stages' (number of separation stages).
    electrolyzer : dict
        Electrolyzer information; must contain the 'type' key. 'type' can equal on of the following: 'Shen'.
        'Shen' requires 'E0' (reversible cell potential) in V, 'K' (Shen et al. 2011 model power coeffecient) in ohm$^{-1}$cm$^{-2}$, 'R' (internal cell resistance) in ohm$^1$cm$^2$, and 'area' (area of the membrane stack) in cm$^2$.
    reactor : dict
        Reactor information; must contain the 'type' key. 'type' can equal on of the following: 'plug_flow_reactor'.
        'plug_flow' requires 'P' (operating pressure of the reactor) in bars and 'T' (operating temperature of the reactor) in Celsius.

    Returns
    -------
    flows : dict
        Dictionary of dictionaries containing the volumetric (L/s) and mass (kg/s) flows for the system.
    power : dict
        Power consumption for each component, and total consumption.
    misc : dict
        Remaining useful non-flow, non-power related information.

    Notes
    -----
    For the input dictionaries, the 'type' value is derived from the function names found in the component's respective submodule.
    """

    # The flow diagram for this system follows:

    # .. image:: ../diagrams/seafuel.png

    # Examples
    # --------


    # degasser
    if degasser['type'] == 'bipolar_membrane':
        mass_co2_out, power_degas = pyseafuel.degasser.bipolar_membrane(degasser['seawater_in'])
        # mass_co2_out; kg/s
        # power_degas; W

        # degasser results dictionary
        degasser_flows = {'seawater_in': degasser['seawater_in'], 'co2_out': mass_co2_out}


    # desalinator
    if desalinator['type'] == 'electrodialysis':
        volume_h2o_out, volume_brine_out, power_desal, concentration_brine = pyseafuel.desalinator.electrodialysis(desalinator['seawater_in'], desalinator['seawater_S'], desalinator['seawater_T'], desalinator['water_ratio'], desalinator['salt_removal'], desalinator['stages'], brine_concentration=True)
        # volume_h2o_out; L/s
        # volume_brine_out; L/s
        # power_desal; W
        # concentration_brine; mol/L

        # desalinator results dictionary
        desalinator_flows = {'seawater_in': desalinator['seawater_in'], 'h2o_out': volume_h2o_out, 'brine_out': volume_brine_out}
        desalinator_misc = {'brine_concentration': concentration_brine}


    # electrolyzer
    if electrolyzer['type'] == 'Shen':
        volume_h2o_in = volume_h2o_out
        mass_h2_out, mass_o2_out, power_electro = pyseafuel.electrolyzer.Shen(volume_h2o_in, electrolyzer['E0'], electrolyzer['K'], electrolyzer['R'], electrolyzer['area'])
        # mass_h2_out; kg/s
        # mass_o2_out; kg/s
        # power_electro; W

        # electrolyzer results dictionary
        electrolyzer_flows = {'h2o_in': volume_h2o_in, 'h2_out': mass_h2_out, 'o2_out': mass_o2_out}


    # reactor
    if reactor['type'] == 'plug_flow':
        mass_co2_in = mass_co2_out
        mass_h2_in = mass_h2_out
        volume_co_out, volume_co2_out, volume_h2_out, volume_h2o_out, volume_ch3oh_out, conversion_factor, _ = pyseafuel.reactor.plug_flow(0, mass_co2_in, mass_h2_in, P=reactor['P'], T=reactor['T'], outflow='volume')
        # volume_co_out; L/s
        # volume_co2_out; L/s
        # volume_h2_out; L/s
        # volume_h2o_out; L/s
        # volume_ch3oh_out; L/s
        # conversion_factor;

        # flow input ratio
        mol_h2_in = mass_h2_out / _constants.mol_weight_h2
        mol_co2_in = mass_co2_out / _constants.mol_weight_co2
        reactor_input_ratio = mol_h2_in / mol_co2_in
        
        # reactor results dictionary
        reactor_flows = {'co2_in': mass_co2_in, 'h2_in': mass_h2_in, 'co_out': volume_co_out, 'co2_out': volume_co2_out, 'h2_out': volume_h2_out, 'h2o_out': volume_h2o_out, 'ch3oh_out': volume_ch3oh_out}
        reactor_misc = {'input_ratio': reactor_input_ratio, 'conversion_factor': conversion_factor}


    # return dictionaries
    
    # flows
    flows = {'degasser': degasser_flows, 'desalinator': desalinator_flows, 'electrolyzer': electrolyzer_flows, 'reactor': reactor_flows}

    # power
    power_total = power_degas + power_desal + power_electro
    power = {'degasser': power_degas, 'desalinator': power_desal, 'electrolyzer': power_electro, 'total': power_total}

    # miscellaneous
    misc = {'desalinator': desalinator_misc, 'reactor': reactor_misc}

    return flows, power, misc



import pyseafuel._constants as _constants
import numpy as np
from scipy import integrate, optimize

# ──────────────────────────────────────────────────────────────────────────

def equilibrium(N_co_0, N_co2_0, N_h2_0, P, T):
    """
    Solves for the equilibrium of the hydrogenation of carbon dioxide reaction.

    Parameters
    ----------
    N_co_0 : float
        Initial amount of carbon monoxide; moles.
    N_co2_0 : float
        Initial amount of carbon dioxide; moles.
    N_h2_0 : float
        Initial amount of hydrogen; moles.
    P : float
        Pressure; bars.
    T : float
        Temperature; Kelvin.

    return N_co, N_co2, N_h2, N_h2o, N_ch3oh, xi_3

    Returns
    -------
    N_co : float
        Equilibrium amount of carbon monoxide; moles.
    N_co2 : float
        Equilibrium amount of carbon dioxide; moles.
    N_h2 : float
        Equilibrium amount of hydrogen; moles.
    N_h2o : float
        Equilibrium amount of water; moles.
    xi_3 : float
        Conversion factor of initial carbon dioxide to methanol.

    Notes
    -----
    This numerical solution for the amounts of reactants/products at equilibrium for the hydrogenation of carbon monoxide follows the method laid out in Terreni et al. 2020[1]_.

    References
    ----------
    .. [1] `J. Terreni, A. Borgschulte, M. Hillestad, and B. Patterson, "Understanding Catalysis-A Simplified Simulation of Catalytic Reactors for CO$_2$ Reduction," ChemEngineering, vol. 4, 2020.<https://doi.org/10.3390/chemengineering4040062>`_
    """

    def _sn(N_co_0, N_co2_0, N_h2_0):
        """"Stoichiometric number" of the initial conditions.
        """

        return (N_h2_0 - N_co2_0)/(N_co_0 + N_co2_0)

    def _F(xi, SN, P, T):
        """Equilibrium equation to solve.
        """

        def _Q_2(xi, SN):
            """"Reaction quotient" of the reverse water gas shift reaction.
            """
            # xi[0] = xi_2
            # xi[1] = xi_3
            return xi[0] * (xi[0] + xi[1]) / ( (SN + 1 - xi[0] - 3*xi[1]) * (1 - xi[0] - xi[1]) )

        def _Q_3(xi, SN, P):
            """"Reaction quotient" of direct hydrogenation of carbon dioxide.
            """

            # xi[0] = xi_2
            # xi[1] = xi_3
            Po = 1.01325  # bars; reference pressure for reduced pressure
            return xi[1] * (xi[0] + xi[1]) * (SN + 2 - 2*xi[1])**2 * Po**2 / ( (SN + 1 - xi[0] - 3*xi[1])**3 * (1 - xi[0] - xi[1]) * P**2 )

        def _K_2(T):
            """Kinetic factor at equilibrium for the reverse water gas shift reaction.
            """
            return 1.07e2 * np.exp( -3.91e4 / (_constants.R * T) )

        def _K_3(T):
            """Kinetic factor at equilibrium for the direct hydrogenation of carbon dioxide.
            """
            return 2.56e-11 * np.exp( 5.87e4 / (_constants.R * T) )

        # return np.abs( np.log(K_2(T)) - np.log( Q_2(xi, SN) ) ) + np.abs( np.log(K_3(T)) - np.log( Q_3(xi, SN, P) ) )
        return [_K_2(T) - _Q_2(xi, SN), _K_3(T) - _Q_3(xi, SN, P)]

    SN = _sn(N_co_0, N_co2_0, N_h2_0)

    # solve for equilibrium conversion coefficients
    xi = optimize.fsolve(_F, [0.1, 0.1], args=(SN, P, T))  # initial values of 0.1 seems to work well for convergence

    # moles of reactants/products at equilibrium
    N_co = xi[0] * N_co2_0
    N_co2 = (1 - xi[0] - xi[1]) * N_co2_0
    N_h2 = (SN + 1 - xi[0] - 3*xi[1]) * N_co2_0
    N_h2o = (xi[0] + xi[1]) * N_co2_0
    N_ch3oh = xi[1] * N_co2_0

    xi_3 = xi[1]  # conversion of initial carbon dioxide to methanol

    return N_co, N_co2, N_h2, N_h2o, N_ch3oh, xi_3


def plug_flow(co_in, co2_in, h2_in, h2o_in=0, ch3oh_in=0, P=60, T=270, rho_catalyst=1000, n_tubes=10000, d_tubes=0.02, L_tubes=3, inflow='mass', outflow='mass', output=None):
    """
    Plug Flow Reactor (PFR) steady state flow simulation of hydrogenation of carbon dioxide to produce methanol with a Cu/ZnO/Al$_2$O$_3$ catalyst[1]_.
    
    Parameters
    ----------
    co_in : float
        Flow rate in of carbon monoxide.
    co2_in : float
        Flow rate in of carbon dioxide.
    h2_in : float
        Flow rate in of hydrogen gas.
    h2o_in : float
        Flow rate in of water.
    ch3oh_in : float
        Flow rate in of methanol.
    P : float (optional)
        Pressure of the reactor; bars; default = 60 bars.
    T : float (optional)
        Temperature of the reactor; Celsius; between [180, 340]; default = 270.
    rho_catalyst : float (optional)
        Density of the catalyst Cu/ZnO/Al:sub:`2`O:sub:`3`; kilograms per cubic meter; default = 1000.
    n_tubes : float (optional)
        Number of tubes of the PFR; default = 10000.
    d_tubes : float (optional)
        Diameter of the PFR tubes; meters; default = 0.02.
    L_tubes : float (optional)
        Length of the tubes; meters; default = 3.
    inflow : string (optional)
        Flow input units; choose between 'mass', 'molar', or 'volume'; kg/s, mol/s, and L/s, respectively; default = 'mass'.
    outflow : string (optional)
        Flow output units; choose between 'mass', 'molar', or 'volume'; kg/s, mol/s, and L/s, respectively; default = 'mass'.
    output : string or None (optional)
        Output flag; if equal to 'total', the entire numerical solution is returned; default = None.

    Returns
    -------
    co_out : float or ndarray
        Carbon monoxide flow rate out; kg/s, mol/s, or L/s.
    co2_out : float or ndarray
        Carbon dioxide flow rate out; kg/s, mol/s, or L/s.
    h2_out : float or ndarray
        Hydrogen gas flow rate out; kg/s, mol/s, or L/s.
    h2o_out : float or ndarray
        Water flow rate out; kg/s, mol/s, or L/s.
    ch3oh_out : float or ndarray
        Methanol flow rate out; kg/s, mol/s, or L/s.
    conversion_factor : float
        Amount of initial carbon dioxide converted to methanol in the reactor.
    x : float or ndarray
        Axial flow coordinate.

    Notes
    -----
    For `outflow = 'volume'`, it is assumed the outflow temperature is cool enough for the water and methanol to condense.

    References
    ----------
    .. [1] `J. Terreni, A. Borgschulte, M. Hillestad, and B. Patterson, "Understanding Catalysis-A Simplified Simulation of Catalytic Reactors for CO$_2$ Reduction," ChemEngineering, vol. 4, 2020.<https://doi.org/10.3390/chemengineering4040062>`_
    """

    def _kinetic_factors(a, b, T):
        """Kinetic factors in Arrhenius form.
        """
        return a * np.exp(b/(_constants.R*T))


    def _dN_dx(x, N_dot, N_dot_co2_0, k, P):
        """Differential equation to solve.
        """
        Po = 1.01325  # bars; reference pressure for reduced pressure

        p = N_dot / np.sum(N_dot) * P / Po  # partial reduced pressures; co, co2, h2, h2o, ch3oh

        # building the reaction rates array
        denom = (1 + k[6]*p[0] + k[7]*p[1]) * (p[2]**.5 + k[8]*p[3])

        r1 = k[3]*k[6]*(p[0]*p[2]**1.5 - p[4]/(p[2]**.5*k[0]))/denom
        r2 = k[4]*k[7]*(p[1]*p[2] - (p[3]*p[0])/k[1])/denom
        r3 = k[5]*k[7]*(p[1]*p[2]**1.5 - (p[4]*p[3])/(p[2]**1.5*k[2]))/denom

        R = np.array([-r1 + r2, -r2 - r3, -2*r1 - r2 - 3*r3, r2 + r3, r1 + r3])

        return R / N_dot_co2_0


    def _dH(T):
        """Enthalpy of formation.
        """
        dH_r1 = -90.5e3 + 42.6*298 - 42.6*T  # J/mol_rxn
        dH_r2 = 41.2e3 + 3.2*298 - 3.2*T  # J/mol_rxn
        dH_r3 = -49.3e3 + 45.8*298 - 45.8*T  # J/mol_rxn

        return np.array([dH_r1, dH_r2, dH_r3])


    def _dH_dx(N_dot_sol, k, P, T):
        """Derivative of the enthalpy of the reactions
        """

        Po = 1.01325  # bars; reference pressure for reduced pressure
        p = N_dot_sol / np.sum(N_dot_sol, axis=0) * P / Po  # partial reduced pressures; co, co2, h2, h2o, ch3oh
        # p = N_dot_sol.y / np.sum(N_dot_sol.y, axis=0) * P / Po  # partial reduced pressures; co, co2, h2, h2o, ch3oh

        # building the reaction rates array; mol/s per kg of catalyst
        denom = (1 + k[6]*p[0] + k[7]*p[1]) * (p[2]**.5 + k[8]*p[3])
        r1 = k[3]*k[6]*(p[0]*p[2]**1.5 - p[4]/(p[2]**.5*k[0]))/denom
        r2 = k[4]*k[7]*(p[1]*p[2] - (p[3]*p[0])/k[1])/denom
        r3 = k[5]*k[7]*(p[1]*p[2]**1.5 - (p[4]*p[3])/(p[2]**1.5*k[2]))/denom

        dH_rj = np.array([r1, r2, r3]) * _dH(T)
        return np.sum(dH_r, axis=0) # W per kg of catalyst


    # Arrhenius Parameters; values in order: K1, K2, K3, k1, k2, k3, K_co, K_co2, K_h2o/K_h2**.5
    a = np.array([2.39e-13, 1.07e2, 2.56e-11, 4.89e7, 9.64e11, 1.09e5, 2.16e-5, 7.05e-7, 6.37e-9])
    b = np.array([9.84e4, -3.91e4, 5.87e4, -1.13e5, -1.53e5, -.875e5, .468e5, .617e5, .840e5])

    T += 273.15  # now in Kelvin
    A_tubes = n_tubes * np.pi * (d_tubes/2)**2  # tube area

    # initial molar flow rates
    if inflow == 'mass':
        N_dot_co_0 = co_in / _constants.mol_weight_co  # mol/s
        N_dot_co2_0 = co2_in / _constants.mol_weight_co2  # mol/s
        N_dot_h2_0 = h2_in / _constants.mol_weight_h2  # mol/s
        N_dot_h2o_0 = h2o_in / _constants.mol_weight_h2o  # mol/s
        N_dot_ch3oh_0 = ch3oh_in / _constants.mol_weight_ch3oh  # mol/s

    elif inflow == 'molar':
        N_dot_co_0 = co_in  # mol/s
        N_dot_co2_0 = co2_in  # mol/s
        N_dot_h2_0 = h2_in  # mol/s
        N_dot_h2o_0 = h2o_in  # mol/s
        N_dot_ch3oh_0 = ch3oh_in  # mol/s

    elif inflow == 'volume':
        print('Not implemented yet')
        return
    else:
        print('Proper inflow option not selected')
        return

    # storing in an array
    N_dot_0 = np.array([N_dot_co_0, N_dot_co2_0, N_dot_h2_0, N_dot_h2o_0, N_dot_ch3oh_0])  # co, co2, h2, h2o, ch3oh

    # nondimensionalizing per kg of catalyst
    N_dot_0 = N_dot_0 / (N_dot_co2_0 * rho_catalyst * A_tubes * L_tubes)

    # calculating kinetic factors
    k = _kinetic_factors(a, b, T)

    # solving numerically
    N_dot_sol = integrate.solve_ivp(_dN_dx, (0, 1), N_dot_0, args=(N_dot_co2_0, k, P), vectorized=True)

    x = N_dot_sol.t  # axial coordinate; [0, 1]

    # solved nondimensionalized per kg of catalyst flows
    N_dot_co = N_dot_sol.y[0]  # 1/kg; carbon monoxide
    N_dot_co2 = N_dot_sol.y[1]  # 1/kg; carbon dioxide
    N_dot_h2 = N_dot_sol.y[2]  # 1/kg; hydrogen
    N_dot_h2o = N_dot_sol.y[3]  # 1/kg; water
    N_dot_ch3oh = N_dot_sol.y[4]  # 1/kg; methanol

    conversion_factor = N_dot_ch3oh[-1] * rho_catalyst * A_tubes * L_tubes  # initial carbon dioxide to methanol conversion ratio

    # nondimensionalized output
    co_out = N_dot_co  # 1/kg
    co2_out = N_dot_co2  # 1/kg
    h2_out = N_dot_h2  # 1/kg
    h2o_out = N_dot_h2o  # 1/kg
    ch3oh_out = N_dot_ch3oh  # 1/kg

    enthalpy = _dH_dx(N_dot_sol.y, k, P, T)
    
    if output != 'total':

        # just outputting the end values
        enthalpy = np.trapz(enthalpy, x=x)
        x = x[-1]
        co_out = co_out[-1]
        co2_out = co2_out[-1]
        h2_out = h2_out[-1]
        h2o_out = h2o_out[-1]
        ch3oh_out = ch3oh_out[-1]

    # determining output units
    if outflow == 'nondim':
        # in per kg of catalyst
        pass

    elif outflow == 'mass':

        co_out *= rho_catalyst * A_tubes * L_tubes * _constants.mol_weight_co  # kg/s
        co2_out *= rho_catalyst * A_tubes * L_tubes * _constants.mol_weight_co2  # kg/s
        h2_out *= rho_catalyst * A_tubes * L_tubes * _constants.mol_weight_h2  # kg/s
        h2o_out *= rho_catalyst * A_tubes * L_tubes * _constants.mol_weight_h2o  # kg/s
        ch3oh_out *= rho_catalyst * A_tubes * L_tubes * _constants.mol_weight_ch3oh  # kg/s

    elif outflow == 'volume':

        co_out *= rho_catalyst * A_tubes * L_tubes *  _constants.R * 273.15 / 1e5 * 1e3  # L/s; flow at STP
        co2_out *= rho_catalyst * A_tubes * L_tubes *  _constants.R * 273.15 / 1e5 * 1e3  # L/s; flow at STP
        h2_out *= rho_catalyst * A_tubes * L_tubes *  _constants.R * 273.15 / 1e5 * 1e3  # L/s; flow at STP
        h2o_out *= rho_catalyst * A_tubes * L_tubes *  _constants.mol_weight_h2o / _constants.rho_h2o * 1e3  # L/s; assuming the outflow is at a low enough temperature that the water has condensed
        ch3oh_out *= rho_catalyst * A_tubes * L_tubes *  _constants.mol_weight_ch3oh / _constants.rho_ch3oh * 1e3  # L/s; assuming the outflow is at a low enough temperature that the methanol has condensed

    elif outflow == 'molar':

        co_out *= rho_catalyst * A_tubes * L_tubes  # mol/s
        co2_out *= rho_catalyst * A_tubes * L_tubes  # mol/s
        h2_out *= rho_catalyst * A_tubes * L_tubes  # mol/s
        h2o_out *= rho_catalyst * A_tubes * L_tubes  # mol/s
        ch3oh_out *= rho_catalyst * A_tubes * L_tubes  # mol/s

    return co_out, co2_out, h2_out, h2o_out, ch3oh_out, conversion_factor, x, enthalpy
